/* Parham Data Products                                                      */
/* Expert Rules Engine                                                       */
/* main/expert/controllers/RestAPI                                           */
/* Copyright 2020 Wayne Parham                                               */

package com.parhamdata.expert.controllers;

import com.parhamdata.expert.model.rulesDetails.expert.ExpertInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.expert.InputData;
import com.parhamdata.expert.model.rulesDetails.expert.OutputData;
import com.parhamdata.expert.model.rulesDetails.loan.LoanDetails;
import com.parhamdata.expert.model.rulesDetails.loan.LoanInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.loan.UserDetails;
import com.parhamdata.expert.model.rulesDetails.promo.PromoInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.promo.ShoppingCart;
import com.parhamdata.expert.model.rulesDetails.promo.PromoItem;
import com.parhamdata.expert.model.rulesDetails.multiPromo.MultiPromoInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.multiPromo.CartItems;
import com.parhamdata.expert.model.rulesDetails.multiPromo.PromoItems;
import com.parhamdata.expert.model.rulesDetails.compoundPromo.CompoundPromoInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.compoundPromo.Cart;
import com.parhamdata.expert.model.rulesDetails.compoundPromo.Promo;
import com.parhamdata.expert.model.rulesDetails.promoAggregate.PromoAggregateInferenceEngine;
import com.parhamdata.expert.model.rulesDetails.promoAggregate.PromoAggregateInput;
import com.parhamdata.expert.model.rulesDetails.promoAggregate.PromoAggregateOutput;
import com.parhamdata.expert.model.rulesEngine.RulesEngine;
import com.parhamdata.expert.services.databaseService.Rule;
import com.parhamdata.expert.services.databaseService.RuleRecord;
import com.parhamdata.expert.services.rulesetFunctions.RulesetFunctions;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.PutMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.http.ResponseEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import jakarta.validation.constraints.NotNull;
import lombok.extern.slf4j.Slf4j;
import java.util.ArrayList;
import java.util.Optional;
import java.util.List;

@Slf4j
@RestController
public class RestAPI {
/* === Rules Engine Connections ======================= */
    @Autowired
    private RulesetFunctions runner;
    @Autowired
    private RulesEngine rulesEngine;
/* === Application Specific Connections =============== */
    @Autowired
    private ExpertInferenceEngine expertInferenceEngine;
    @Autowired
    private LoanInferenceEngine loanInferenceEngine;
    @Autowired
    private PromoInferenceEngine promoInferenceEngine;
    @Autowired
    private MultiPromoInferenceEngine multiPromoInferenceEngine;
    @Autowired
    private CompoundPromoInferenceEngine compoundPromoInferenceEngine;
    @Autowired
    private PromoAggregateInferenceEngine aggregatePromoInferenceEngine;

/* === Application Specific Mappings ================== */
    @PostMapping(value = "/loan")
    public ResponseEntity<?> postUserLoanDetails(@RequestBody UserDetails userDetails) {
        log.info("Requesting analysis about loan for user: " + userDetails);
        LoanDetails result = (LoanDetails) rulesEngine.run(loanInferenceEngine, userDetails, false);
        log.info("Result: " + result);
        if (result == null) {
            result = LoanDetails.builder()
                    .notes("No matching rule.  Please examine input data and ruleset.")
                    .build();
        }
        return ResponseEntity.ok(result);
    }

    @PostMapping(value = "/promo")
    public ResponseEntity<?> postShoppingCart(@RequestBody ShoppingCart cart) {
        log.info("Requesting analysis for cart id # " + cart.getId());
        PromoItem result = (PromoItem) rulesEngine.run(promoInferenceEngine, cart, false);
        log.info("Result: " + result);
        if (result == null) {
            result = PromoItem.builder()
                    .notes("No matching rule.  Please examine input data and ruleset.")
                    .build();
        }
        return ResponseEntity.ok(result);
    }

    @PostMapping(value = "/multiPromo")
    public ResponseEntity<?> postMultiPromo(@RequestBody CartItems cart) {
        log.info("Requesting analysis for cart id # " + cart.getId());
        List<PromoItems> response = (List<PromoItems>) rulesEngine.run(multiPromoInferenceEngine, cart, true);
        log.info("Result: " + response);
        if (response == null) {
            String errorMessage = "{\n\t\"status\": false,";
            errorMessage += "\n\t\"notes\": \" No matching rule.  Please examine input data and ruleset.\"\n}";
            HttpHeaders responseHeaders = new HttpHeaders();
            responseHeaders.add("Content-Type", "application/json");
            return new ResponseEntity<>(errorMessage, responseHeaders, HttpStatus.OK);
        }
        else {
            return ResponseEntity.ok(response);
        }
    }

    @PostMapping(value = "/compoundPromo")
    public ResponseEntity<?> postCompoundPromo(@RequestBody Cart cart) {
        log.info("Requesting analysis for cart id # " + cart.getId());
        List<Promo> prelims = (List<Promo>) rulesEngine.run(compoundPromoInferenceEngine, cart, true);
        log.info("Preliminary result: " + prelims);
        ArrayList<PromoAggregateInput.Promo> availablePromos = new ArrayList<>();
        for(Promo prelim : prelims) {
            PromoAggregateInput.Promo inputPromo = new PromoAggregateInput.Promo();
            inputPromo.setStatus(prelim.getStatus());
            inputPromo.setNotes(prelim.getNotes());
            inputPromo.setCartTotal(prelim.getCartTotal());
            inputPromo.setPromoUpc(prelim.getPromoUpc());
            inputPromo.setPromoDescription(prelim.getPromoDescription());
            inputPromo.setPromoCategory(prelim.getPromoCategory());
            inputPromo.setPromoPrice(prelim.getPromoPrice());
            inputPromo.setPromoQty(prelim.getPromoQty());
            availablePromos.add(inputPromo);
        }
        PromoAggregateInput aggregate = new PromoAggregateInput(availablePromos);
        PromoAggregateOutput response = (PromoAggregateOutput) rulesEngine.run(aggregatePromoInferenceEngine, aggregate, false);
        log.info("Final result: " + response);
        if (response == null) {
            response = PromoAggregateOutput.builder()
                    .notes("No matching rule.  Please examine input data and ruleset.")
                    .build();
        }
        return ResponseEntity.ok(response);
    }

    @PostMapping(value = "/zork")
    public ResponseEntity<?> postZorkPayload(@RequestBody InputData inputData) {
        log.info("Requesting analysis for input: " + inputData);
        OutputData result = (OutputData) rulesEngine.run(expertInferenceEngine, inputData, false);
        log.info("Result: " + result);
        if (result == null) {
            result = OutputData.builder()
                    .result("You have fallen off the edge of the world.")
                    .location("unknown")
                    .inventory("nothing")
                    .status("dead")
                    .build();
        }
        return ResponseEntity.ok(result);
    }

    /* === Rules Engine Mappings ========================== */
    @GetMapping(value = "/getRules/{domain}")
    public ResponseEntity<?> getRulesInDomain(@NotNull @PathVariable("domain") String domain) {
        log.info("Requesting rulesets for '" + domain.toUpperCase() + "'." );
        List<Rule> Rules = runner.getAllRulesInDomain(domain.toUpperCase());
        log.info("Found " + Rules.size() + " rules in the " + domain.toUpperCase() + " domain." );
        return ResponseEntity.ok(Rules);
    }

    @GetMapping(value = "/getRules")
    public ResponseEntity<?> getAllRules() {
        log.info("Requesting all rulesets.");
        List<Rule> allRules = runner.getAllRules();
        log.info("Found " + allRules.size() + " rules in all domains." );
        return ResponseEntity.ok(allRules);
    }

    @GetMapping(value = "/getRule/{id}")
    public ResponseEntity<?> getRuleById(@NotNull @PathVariable("id") Integer id) {
        log.info("Requesting rules with ID [" + id + "] in any domain." );
        List<Rule> Rules = runner.getRuleById(id);
        log.info("Result: " + Rules);
        return ResponseEntity.ok(Rules);
    }

    @GetMapping(value = "/getRule/{domain}/{id}")
    public ResponseEntity<?> getRule(@NotNull @PathVariable String domain, @NotNull @PathVariable Integer id) {
        log.info("Requesting rule [" + domain.toUpperCase() + "][" + id + "].");
        Optional<RuleRecord> Rule = runner.getRule(domain.toUpperCase(), id);
        log.info("Result: " + Rule);
        return ResponseEntity.ok(Rule);
    }

    @PostMapping("/setRule")
    public RuleRecord postWrite (@RequestBody RuleRecord rule) {
        log.info("Setting rule [" + rule.getDomain() + "][" + rule.getId() + "].");
        RuleRecord result = runner.save(rule);
        log.info("Record: " + result.toString());
        return result;
    }

    @PutMapping("/setRule")
    public RuleRecord putWrite (@RequestBody RuleRecord rule) {
        log.info("Setting rule [" + rule.getDomain() + "][" + rule.getId() + "].");
        RuleRecord result = runner.save(rule);
        log.info("Record: " + result.toString() );
        return result;
    }

    @PostMapping("/loadRules")
    public List<RuleRecord> postLoad (@RequestBody List<RuleRecord> rules) {
        log.info("Uploading " + rules.size() + " rules.");
        List<RuleRecord> results = runner.bulkLoad(rules);
        log.info("Sucessfully uploaded " + results.size() + " rules.");
        return results;
    }

    @PutMapping("/loadRules")
    public List<RuleRecord> putLoad (@RequestBody List<RuleRecord> rules) {
        log.info("Uploading " + rules.size() + " rules.");
        List<RuleRecord> results = runner.bulkLoad(rules);
        log.info("Sucessfully uploaded " + results.size() + " rules.");
        return results;
    }

    @DeleteMapping("/deleteRule/{id}")
    public ResponseEntity<?> deleteExpertRule(@PathVariable Integer id) {
        log.info("Deleting rule [EXPERT][" + id + "].");
        List<RuleRecord> deletedRules = runner.deleteByDomainAndId("", id);
        if (deletedRules.size() == 0) {
            log.info("Rule [EXPERT][" + id.toString() + "] not found.");
        }
        else {
            log.info("Deleted rule [EXPERT][" + id.toString() +  "].");
        }
        return ResponseEntity.ok(deletedRules);
    }

    @DeleteMapping("/deleteRule/{domain}/{id}")
    public ResponseEntity<?> deleteDomainRule(@PathVariable String domain, @PathVariable Integer id) {
        log.info("Deleting rule [" + domain.toUpperCase() + "][" + id.toString() +  "].");
        List<RuleRecord> deletedRules = runner.deleteByDomainAndId(domain.toUpperCase(), id);
        if (deletedRules.size() == 0) {
            log.info("Rule [" + domain.toUpperCase() + "][" + id + "] not found.");
        }
        else {
            log.info("Deleted rule [" + domain.toUpperCase() + "][" + id +  "].");
        }
        return ResponseEntity.ok(deletedRules);
    }

    @DeleteMapping("/deleteDomain/{domain}")
    public ResponseEntity<?> delete(@PathVariable String domain) {
        log.info("Deleting domain '" + domain.toUpperCase() + "'.");
        List<RuleRecord> deletedRules = runner.deleteByDomain(domain.toUpperCase());
        log.info("Deleted " + deletedRules.size() + " rules from the '" + domain.toUpperCase() + "' domain." );
        return ResponseEntity.ok(deletedRules);
    }

    @DeleteMapping("/deleteAllRules")
    public ResponseEntity<String> delete() {
        log.info("Deleting all rules.");
        runner.deleteAll();
        String response = "[]";
        HttpHeaders responseHeaders = new HttpHeaders();
        responseHeaders.add("Content-Type", "application/json");
        return new ResponseEntity<>(response, responseHeaders, HttpStatus.OK);
    }
}