/* Parham Data Products                                                      */
/* Expert Rules Engine                                                       */
/* main/expert/model/rulesEngine/langParser/VariableParser                   */
/* Copyright 2020 Wayne Parham                                               */

package com.parhamdata.expert.model.rulesEngine.langParser;

import com.parhamdata.expert.model.rulesEngine.variableResolver.VariableResolver;
import com.parhamdata.expert.model.rulesEngine.variableResolver.KeywordResolver;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import lombok.extern.slf4j.Slf4j;
import java.util.stream.Collectors;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Slf4j
@Service
public class VariableParser {

    @Autowired
    private KeywordResolver keywordResolver;
    @Autowired
    private KeywordParser keywordParser;

    public String resolveKeywords(String expression) {
        Map<String, Object> keywordToResolverValueMap = executeKeywordResolver(expression);
        return replaceKeywordsWithValue(expression, keywordToResolverValueMap);
    }

    private Map<String, Object> executeKeywordResolver(String expression) {
        List<String> listOfKeywords = keywordParser.getListOfKeywords(expression);
        Map<String, Object> keywordToResolverValueMap = new HashMap<>();
        listOfKeywords.stream()
                .forEach(
                        keyword -> {
                            String extractedKeyword = keywordParser.extractKeyword(keyword);
                            String keyResolver = keywordParser.getKeywordResolver(extractedKeyword);
                            String keywordValue = keywordParser.getKeywordValue(extractedKeyword);
                            VariableResolver resolver = keywordResolver.getResolver(keyResolver).get();
                            Object resolveValue = resolver.resolveValue(keywordValue);
                            keywordToResolverValueMap.put(keyword, resolveValue);
                        }
                );
        return keywordToResolverValueMap;
    }

    private String replaceKeywordsWithValue(String expression, Map<String, Object> keywordToResolverValueMap) {
        String expr = expression;
        List<String> keyList = keywordToResolverValueMap.keySet().stream().collect(Collectors.toList());
        for (int index = 0; index < keyList.size(); index++) {
            try {
                String key = keyList.get(index);
                String resolveValue = keywordToResolverValueMap.get(key).toString();
                expression = expression.replace(key, resolveValue);
                }
            catch (Exception e) {
                log.error("Error in expression: [" + expr + "]");
                return null;
            }
        }
        return expression;
    }
}